from .helpers import get_regex_images
import numpy as np
from skimage.morphology._util import _offsets_to_raveled_neighbors
import torch
import torch.nn as nn


def channel_losses_to_dict(inputs, targets, channels, loss, loss_dict):
    #if targets.max() < 1 and targets.min() > 0:
    #loss = nn.BCELoss()
    #else:
       # loss = nn.MSELoss()
    for i, c in enumerate(channels):
        c_input = inputs[:, i, ...].detach()
        c_target = targets[:, i, ...].detach()
        l = loss(c_input, c_target)
        loss_dict[c].append(l.item())


# --------------
# Loss Functions
# --------------

class DiceLoss(nn.Module):
    '''
    DiceLoss: 1 - DICE coefficient 

    Adaptations: weights output channels equally in final loss. 
    This is necessary for anisotropic data.
    '''
    def __init__(self, weight=None, size_average=True):
        super(DiceLoss, self).__init__()

    def forward(self, inputs, targets, channel_dim=1, smooth=1):
        '''
        inputs: torch.tensor
            Network predictions. Float
        targets: torch.tensor
            Ground truth labels. Float
        channel_dim: int
            Dimension in which output channels can be found.
            Loss is weighted equally between output channels.
        smooth: int
            Smoothing hyperparameter.
        '''
        
        #comment out if your model contains a sigmoid or equivalent activation layer
        #inputs = F.sigmoid(inputs) 
        inputs, targets = flatten_channels(inputs, targets, channel_dim)
        intersection = (inputs * targets).sum(-1) 
        dice = (2.*intersection + smooth)/(inputs.sum(-1) + targets.sum(-1) + smooth) 
        loss = 1 - dice 
        return loss.mean()


class WeightedBCELoss(nn.Module):
    def __init__(
            self, 
            chan_weights, 
            device,
            reduction='mean', 
            final_reduction='mean', 
            channel_dim=1,
        ):
        super(WeightedBCELoss, self).__init__()
        self.bce = nn.BCELoss(reduction='none')
        self.chan_weights = torch.tensor(list(chan_weights)).to(device)
        self.reduction = reduction
        self.final_reduction = final_reduction
        self.channel_dim = channel_dim
        self.device = device


    def forward(self, inputs, targets):
        loss = weighted_BCE_loss(
            inputs, 
            targets, 
            self.chan_weights, 
            self.device,
            self.channel_dim, 
            self.reduction, 
            self.final_reduction
        )
        return loss



class EpochwiseWeightedBCELoss(nn.Module):
    def __init__(
            self, 
            weights_list, 
            device, 
            reduction='mean', 
            final_reduction='mean', 
            channel_dim=1,
        ):
        super(EpochwiseWeightedBCELoss, self).__init__()
        self._weights = torch.tensor(weights_list, requires_grad=False).to(device)
        self._reduction = reduction
        self._final_reduction = final_reduction
        self._channel_dim = channel_dim
        self._current_epoch = None
        self.current_weights = None
        self.device = device 

    
    @property
    def current_epoch(self):
        return self._current_epoch
    

    @current_epoch.setter
    def current_epoch(self, e):
        self._current_epoch = e
        self.current_weights = self._weights[e]


    def forward(self, inputs, targets):
        loss = weighted_BCE_loss(
            inputs, 
            targets, 
            self.current_weights, 
            self.device,
            self._channel_dim, 
            self._reduction, 
            self._final_reduction
        )
        return loss


class ChannelwiseLoss(torch.nn.Module):
    def __init__(self, losses, channels, device, channel_dim=1, 
                 ndims=5, reduction='mean'):
        super(ChannelwiseLoss, self).__init__()
        self.losses = [loss.to(device) for loss in losses]
        self.reduction = reduction
        slices = []
        for c in channels:
            s_ = [slice(None, None),] * ndims
            s_[channel_dim] = c
            s_ = tuple(s_)
            slices.append(s_)
        self.slices = slices

    def forward(self, inputs, targets):
        losses = torch.zeros(len(self.losses))
        for i, s_ in enumerate(self.slices):
            ipt = inputs[s_]
            tgt = targets[s_]
            loss = self.losses[i](ipt, tgt)
            losses[i] = loss
        if self.reduction == 'mean':
            loss = losses.mean()
        elif self.reduction == 'sum':
            loss = losses.sum()
        else:
            raise ValueError(f'{self.reduction} is not a valid reduction for this loss')
        return loss


def weighted_BCE_loss(
        inputs, 
        targets, 
        chan_weights, 
        device,
        channel_dim=1, 
        reduction='mean', 
        final_reduction='mean'
    ):
    bce = nn.BCELoss(reduction='none').to(device)
    inputs, targets = flatten_channels(inputs, targets, channel_dim)
    unreduced = bce(inputs, targets)
    if reduction == 'mean':
        channel_losses = unreduced.mean(-1) * chan_weights
    elif reduction == 'sum':
        channel_losses = unreduced.sum(-1) * chan_weights
    else:
        raise ValueError('reduction param must be mean or sum')
    if final_reduction == 'mean':
        loss = channel_losses.mean()
    elif final_reduction == 'sum':
        loss = channel_losses.sum()
    else:
        raise ValueError('final_reduction must be mean or sum')
    return loss


class SegmentationLoss(nn.Module):

    '''
    Here's the thing... this is super difficult because it can't be used to train a network
    with backprop (i.e., segmentation process isn't differentiable). Would require another
    means of training weights.
    '''
    def __init__(self, GT_dir, ids, seg_func, seg_kwargs):
        super(SegmentationLoss, self).__init__()
        self._func = seg_func 
        self._kwargs = seg_kwargs
        self._batch_no = None
        # get the GTs in correct order
        self._GT_stack = get_regex_images(GT_dir, r'\d{6}_\d{6}_\d{1,3}_GT.tif', ids)


    def forward(self, inputs, targets):
        pass


    def segment(self, data):
        return self._func(data, **self._kwargs)


    def get_GT(self):
        array = self._GT_stack[self._batch_no].compute()
        return array


    @property
    def batch_no(self):
        return self._batch_no


    @batch_no.setter
    def batch_no(self, i):
        self._batch_no = i


# ----------------
# Helper Functions 
# ----------------

def flatten_channels(inputs, targets, channel_dim):
    '''
    Helper function to flatten inputs and targets for each channel

    E.g., (1, 3, 10, 256, 256) --> (3, 655360)

    Parameters
    ----------
    inputs: torch.Tensor
        U-net output
    targets: torch.Tensor
        Target labels
    channel_dim: int
        Which dim represents output channels? 
    '''
    order = [channel_dim, ]
    for i in range(len(inputs.shape)):
        if i != channel_dim:
            order.append(i)
    inputs = inputs.permute(*order)
    inputs = torch.flatten(inputs, start_dim=1)
    targets = targets.permute(*order)
    targets = torch.flatten(targets, start_dim=1)
    return inputs, targets

  
# -------------------
# Probably Not Useful
# -------------------
 
class BCELossWithCentrenessPenalty(nn.Module):

    def __init__(
                 self, 
                 weight=0.1
                 ):
        super(BCELossWithCentrenessPenalty, self).__init__()
        self.weight = weight
        self.bce = nn.BCELoss()

    
    def forward(self, inputs, targets):
        loss = self.bce(inputs, targets)
        affinities = inputs[:3, ...].detach()
        centreness = targets[3, ...].detach()
        penalty = self.centreness_penalty(affinities, centreness)
        loss = loss.subtract(self.weight * penalty)
        return loss


    def centreness_penalty(self, affinities, centreness):
        '''
        Parameters
        ----------
        affinities: torch.Tensor
            shape (3, z, y, x)
        centreness: torch.Tensor
            shape (z, y, x)
        '''
        scores = []
        for i in range(affinities.shape[0]):
            aff = affinities[i].detach()
            score = self.bce(aff, centreness)
            scores.append(score)
        scores = torch.tensor(scores)
        score = scores.mean()
        return score



class CentroidLoss(nn.Module):

    def __init__(
                 self, 
                 selem=np.ones((3, 9, 9)), 
                 centre=(4, 4, 4), 
                 chan_weights=(1., 1., 1.), 
                 eta = 0.5,
                 phi = 0.5
                 ):
        super(WeightedBCELoss, self).__init__()
        self.selem = selem
        self.centre = centre
        self.shape = None
        self.BCELoss = WeightedBCELoss(chan_weights=chan_weights)
        self.eta = eta
        self.phi = phi


    def forward(
                self, 
                inputs, 
                targets, 
                selem=np.ones((3, 9, 9)), 
                centre=(1, 4, 4), 
                channel_dim=1
                ):
        
        # Prep, because flat is easier
        block_shape = [s for s in inputs.shape[-3:]]
        block_shape = tuple(block_shape)
        inputs, targets = flatten_channels(inputs, targets, self.channel_dim)

        # -----------------------
        # BCE Loss for Affinities
        # -----------------------
        # ???? pytorch will make this unexpectedly difficult I'm sure
        input_affs = inputs[:3] # pretty sure this won't work LOL
        target_affs = targets[:3]
        loss = self.BCELoss(input_affs, target_affs)

        # -------------------
        # Centroid Similarity
        # -------------------
        target_cent = targets[3].numpy()
        # penalise similarity of affinities to centroids - high score
        aff_penalties = []
        for i in range(3):
            aff_chan = inputs[i].numpy()
            penalty = self.centroid_similarity(aff_chan, target_cent)
        aff_penalties = np.array(aff_penalties)
        aff_penalty = aff_penalties.mean() * self.phi
        aff_penalties = torch.from_numpy(aff_penalty)
        # penalise difference of centroid output channel from centroids
        input_cent = inputs[3].numpy()
        cent_penalty = (1 - self.centroid_similarity(input_cent, target_cent)) * self.eta
        cent_penalty = torch.from_numpy(cent_penalty)
        # add penalties to 
        loss.add_(aff_penalty)
        loss.add_(cent_penalty)
        return loss



    def centroid_similarity(
                            self, 
                            inputs, 
                            targets,
                            shape 
                            ):
        """
        Metric for similarity to centroid. Computes a similarity based on
        inverse of normaised euclidian distance for every centroid neighbor.
        Neighbors are determined by a structing element (cell-ish dims)

        The output should be between 0-1 (therefore easily invertable)
        MAKE SURE THIS IS TRUE!!!

        Notes
        -----
        Currently uses a structing element to find centroid neighbors.
        Hoping to add an option for using segmentation to get neighbors. 

        """
        offsets = _offsets_to_raveled_neighbors(shape, 
                                                self.selem, 
                                                self.centre)

        euclid_dists = self.euclidian_distances()
        weights = euclid_dists - 1
        centroids = np.argwhere(targets == 1.)
        score = 0
        for c in centroids:
            max_ind = inputs.shape[-1]
            raveled_indices = c + offsets
            in_bounds_indices = np.array([idx for idx in raveled_indices \
                                            if idx >= 0 and idx < max_ind])
            neighbors = inputs[in_bounds_indices]
            weighted = neighbors * weights
            score += weighted.mean()
        return score


    def euclidian_distances(self):
        '''
        Compute euclidian distances of each index from 
        '''
        selem_indices = np.stack(np.nonzero(self.selem), axis=-1)
        distances = []
        centre = np.array(self.centre)
        for ind in selem_indices:
            dist = np.linalg.norm(centre - ind)
            distances.append(dist)
        distances = np.array(distances)
        return distances / distances.max()




import numpy as np
import torch 
import torch.nn as nn
import torch.nn.functional as F


# globals
decoder_instructions = {
            5: {
                'in' : 256 * 2, 
                'out': 128
            }, 
            6: {
                'in' : 128 * 2, 
                'out': 64
            }, 
            7: {
                'in' : 64 * 2, 
                'out': 32
            }, 
        }


# convolution module
class ConvModule(nn.Module):
    def __init__(
                 self, 
                 in_channels, 
                 out_channels, 
                 kernel_size=3, 
                 stride=1, 
                 padding0=1, 
                 padding1=1, 
                 final='relu'
                 ):
        """
        Convolution module with 2x batch normaisation --> convolution. 
        First activation function is ReLU. Second activation defaults to ReLU
        but options also include SoftMax and Sigmoid.

        Parameters
        ----------
        in_channels: int 
            Number of input channels for data
        out_channels: int
            Number of desired output channels
        kernel_size: int or tuple of int
            Size of convolution kernel
        stride: int or tuple of int
            Size of stride to use during convolution
        padding0: int or tuple of int
            Size with which to pad data for first convolution
        padding1: int or tuple of int
            Size with which to pad data for second convolution
        final: str
            Option for final activation function: 
            'relu', 'softmax', or 'sigmoid'
        """
        super(ConvModule, self).__init__()

        # Convolutions
        # ------------
        self.conv0 = nn.Conv3d(
                               in_channels, 
                               out_channels, 
                               kernel_size=kernel_size, 
                               stride=stride, 
                               padding=padding0
                               )
        self.conv1 = nn.Conv3d(
                               out_channels, 
                               out_channels, 
                               kernel_size=kernel_size, 
                               stride=stride, 
                               padding=padding1
                               )

        # Batch Normailsation
        # -------------------
        self.batch0 = nn.BatchNorm3d(out_channels)
        self.batch1 = nn.BatchNorm3d(out_channels)

        # Activation
        # ----------
        self.relu0 = nn.ReLU()
        self.relu1 = nn.ReLU()
        self.sm = nn.Softmax()
        self.final = final


    def forward(self, x):
        # First convolution
        x = self.conv0(x)
        x = self.batch0(x)
        x = self.relu0(x)
        x = self.conv1(x)
        x = self.batch1(x)
        if self.final == 'relu':
            x = self.relu1(x)
        elif self.final == 'softmax':
            x = self.sm(x)
        elif self.final == 'sigmoid':
            x = torch.sigmoid(x)
        elif self.final == 'tanh':
            x = torch.tanh(x)
        return x


class ResConvModule(nn.Module):

    def __init__(
                 self, 
                 in_channels, 
                 out_channels, 
                 kernel_size=3, 
                 stride=1, 
                 padding0=1, 
                 padding1=1, 
                 final='relu'):
        """
        """
        super(ResConvModule, self).__init__()


# trial UNet
class UNet(nn.Module):

    def __init__(
                 self, 
                 in_channels=1, 
                 out_channels=5, 
                 down_factors=(1, 2, 2), 
                 up='convolution', 
                 downsample_1_at_bottom=True, 
                 chan_final_activations=None
                 ):
        '''
        Anisotropic U-net

        Parameters
        ----------
        in_channels: int
        out_channels: int
        down_factors: tuple of int
            Factors by which to downsample in encoder
        up: str
            'bilinear': use bilinear/nearest neighbour interpolations for up 
            sampling in decoder
            'convolution': use inverse convolutions with learnable parameters
        '''
        super(UNet, self).__init__()
        self.forked = isinstance(out_channels, tuple)
        if not self.forked:
            out_channels = (out_channels,)
        else:
            self.forks = len(out_channels)
        self.out_channels = out_channels
        # Max pooling 
        # -----------
        # encoder: downsample 4 times
        # Get the padding for the max pool
        #   Must be at most half of 
        p = [np.floor_divide(df, 2).astype(int) for df in down_factors]
        p = tuple(p)
        # max pool layers
        self.d0 = nn.MaxPool3d(
                               down_factors, 
                               stride=down_factors, 
                               padding=(0, 1, 1)
                               )
        self.d1 = nn.MaxPool3d(
                               down_factors, 
                               stride=down_factors, 
                               padding=p
                               )
        self.d2 = nn.MaxPool3d(
                               down_factors, 
                               stride=down_factors, 
                               padding=(0, 1, 1)
                               )
        # the input will be downsampled in previously un
        new_down = self.new_down_factors(down_factors, downsample_1_at_bottom)
        self.d3 = nn.MaxPool3d(
                               new_down, 
                               stride=new_down, 
                               padding=(0, 1, 1)
                               )

        # Convolutions
        # ------------
        # encoder colvolutions:
        self.c0 = ConvModule(in_channels, 32)
        self.c1 = ConvModule(32, 64)
        self.c2 = ConvModule(64, 128)
        self.c3 = ConvModule(128, 256)
        self.c4 = ConvModule(256, 256)

        # decoder convolutions
        for i, c in enumerate(out_channels):
            for key in decoder_instructions.keys():
                n, in_c, = key, decoder_instructions[key]['in']
                out_c = decoder_instructions[key]['out']
                cmd = f'self.c{n}_{i} = ConvModule({in_c}, {out_c})'
                exec(cmd)
            if chan_final_activations is not None:
                final = chan_final_activations[i]
            else:
                final = 'sigmoid'
            cmd = f'self.c8_{i} = ConvModule(32 * 2, {c}, final=\'{final}\')'
            exec(cmd)

        # Upsampling
        # ----------
        # Inverse convolutions
        if up == 'convolution':
            self.up0 = nn.ConvTranspose3d(
                                          256, 
                                          256, 
                                          kernel_size=new_down, 
                                          stride=new_down, 
                                          groups=256)
            self.up1 = nn.ConvTranspose3d(
                                          128, 
                                          128, 
                                          kernel_size=down_factors, 
                                          stride=down_factors, 
                                          groups=128
                                          )
            self.up2 = nn.ConvTranspose3d(
                                          64, 
                                          64, 
                                          kernel_size=down_factors, 
                                          stride=down_factors, 
                                          groups=64
                                          )
            self.up3 = nn.ConvTranspose3d(
                                          32, 
                                          32, 
                                          kernel_size=down_factors, 
                                          stride=down_factors, 
                                          groups=32
                                          )
        elif up == 'bilinear':
            self.up0 = lambda x: self.bilinear_interpolation(x, down_factors)
            self.up1 = lambda x: self.bilinear_interpolation(x, down_factors)
            self.up2 = lambda x: self.bilinear_interpolation(x, down_factors)
            self.up3 = lambda x: self.bilinear_interpolation(x, down_factors)
        else:
            raise ValueError('Valid options for up param are convolution and bilinear')


    def bilinear_interpolation(self, x, down_factors):
        '''
        THIS NEEDS DEBUGGING!!
        '''
        print(x.shape)
        x = torch.squeeze(x, 0)
        print(x.shape)
        x = F.interpolate(
                          x, 
                          mode='tconv', 
                          scale_factor=down_factors
                          )
        x = torch.unsqueeze(x, 0)
        return x


    def new_down_factors(self, down_factors, downsample_1_at_bottom):
        if downsample_1_at_bottom:
            # downsample in last pool layer if down factor has been 1
            # for an axis
            new_down = []
            for df in down_factors:
                if df == 1:
                    new_down.append(2)
                else:
                    new_down.append(df)
            new_down = tuple(new_down) # probs not necessary, I like it, sue me
        else:
            new_down = down_factors
        return new_down


    def forward(self, x):
        # Encoder
        # -------
        x, c0, c1, c2, c3 = self.encoder(x)
 
        # Decoder
        # -------
        if self.forked:
            x = self.forked_decoder(x, c0, c1, c2, c3)
        else:
            x = self.decoder(x, c0, c1, c2, c3)
        return x

    
    def encoder(self, x):
        # Encoder
        # -------
        c0 = self.c0(x)
        x = self.d0(c0)
        c1 = self.c1(x)
        x = self.d1(c1)
        c2 = self.c2(x)
        x = self.d2(c2)
        c3 = self.c3(x)
        x = self.d3(c3)
        x = self.c4(x)
        return x, c0, c1, c2, c3


    def forked_decoder(self, x, c0, c1, c2, c3):
        # Decoder
        # -------
        started = False
        for i in range(self.forks):
            x = x.clone()
            if not started:
                x0 = self.decoder(x, c0, c1, c2, c3, i=i)
                started = True
            else:
                x1 = self.decoder(x, c0, c1, c2, c3, i=i)
                x0 = torch.cat((x0, x1), dim=1)
        return x0


    def decoder(self, x, c0, c1, c2, c3, i=0):
        x = self.up0(x)
        # quick dumb hack for concatenation 
        x = x[:, :, :, :-1, :-1]
        x = torch.cat([x, c3], 1)
        if i == 0:
            x = self.c5_0(x)
            x = self.up1(x)
            x = x[:, :, :, :-1, :-1]
            x = torch.cat([x, c2], 1)
            x = self.c6_0(x)
            x = self.up2(x)
            x = x[:, :, :, :-1, :-1]
            x = torch.cat([x, c1], 1)
            x = self.c7_0(x)
            x = self.up3(x)
            x = x[:, :, :, 1:-1, 1:-1]
            x = torch.cat([x, c0], 1)
            x = self.c8_0(x)
        elif i == 1:
            x = self.c5_1(x)
            x = self.up1(x)
            x = x[:, :, :, :-1, :-1]
            x = torch.cat([x, c2], 1)
            x = self.c6_1(x)
            x = self.up2(x)
            x = x[:, :, :, :-1, :-1]
            x = torch.cat([x, c1], 1)
            x = self.c7_1(x)
            x = self.up3(x)
            x = x[:, :, :, 1:-1, 1:-1]
            x = torch.cat([x, c0], 1)
            x = self.c8_1(x)
            # so on and so forth? Couldn't make the below work
        #cmd = f'x = self.c8_{i}(x)'
        #exec(cmd)
        return x

    




class ForkedUNet(UNet):
    def __init__(self, in_channels=1, fork_channels=(8, 2)):
        super(ForkedUNet, self).__init__(in_channels=in_channels, out_channels=fork_channels)
        self.forks = len(fork_channels)


    def forward(self, x):
        # Encoder
        # -------
        x, c0, c1, c2, c3 = self.encoder(x)
        print('forked unet')
 
        # Decoder
        # -------
        started = False
        for i in range(self.forks):
            x = x.clone()
            if not started:
                x0 = self.decoder(x, c0, c1, c2, c3, i=i)
                print('out shape: ', x0.shape)
            else:
                x1 = self.decoder(x, c0, c1, c2, c3, i=i)
                print('out shape: ', x1.shape)
                x0 = torch.cat(x0, x1, dim=1)
        return x0



if __name__ == '__main__':
    ip = torch.randn(1, 1, 10, 256, 256)
    unet = UNet()
    o = unet(ip)
    # output has shape (1, 3, 10, 256, 256)

    # c0 -> torch.Size([1, 32, 10, 256, 256])
    # d0, c1 -> torch.Size([1, 64, 10, 129, 129])
    # d1, c2 -> torch.Size([1, 128, 10, 65, 65])
    # d2, c3 -> torch.Size([1, 256, 10, 33, 33])
    # d3, c4 -> torch.Size([1, 256, 10, 17, 17]
    # u0 -> torch.Size([1, 256, 10, 33, 33])
    # u1, c5 -> torch.Size([1, 128, 10, 65, 65])
    # u2, c6 -> torch.Size([1, 64, 10, 129, 129])
    # u3, c7 -> torch.Size([1, 32, 10, 256, 256])
    # c8 -> torch.Size([1, 3, 10, 256, 256])


from datetime import datetime
import os
from . import train
from .train_io import get_train_data
import torch.nn as nn
from pathlib import Path


def run_experiment(
    experiment_dict, 
    image_list, 
    labels_list, 
    out_dir, 
    *args, 
    **kwargs
    ):
    gtd_kwargs = experiment_dict['get_train_data']
    train_dict = get_train_data(image_list, labels_list, out_dir, **gtd_kwargs)
    unets = {}
    for key in train_dict.keys():
        train_kwargs = train_dict[key]
        train_kwargs.update(experiment_dict[key])
        unet, unet_path = train.train_unet(**train_kwargs)
        unets[key] = {'unet': unet, 'unet_path' : unet_path}
    upper_dir = Path(train_kwargs['out_dir']).parents[1]
    unet_path_log = upper_dir / 'unet_paths.txt'
    s = [unets[key]['unet_path'] for key in unets.keys()]
    with open(unet_path_log, 'a') as f:
        st = str(s)
        f.write(st)
    return s


#def train_experiment(
 #   experiments, 
  #  out_dir, 
   # image_paths, 
#    labels_path
 #   exp, 
  #  ):
   # exp_dir = os.path.join(out_dir, experiments[exp]['suffix'])
    #    unet = train.train_unet_get_labels(
     #           exp_dir, 
      #          image_paths, 
       #         labels_paths, 
        #        **experiments[exp]) 
       # exp['unet'] = unet


def get_experiment_dict(
    channels_list,
    condition_names,
    conditions_list=None,
    name='train-unet' ,
    validation_prop=0.2, 
    n_each=100,
    scale=(4, 1, 1),
    **kwargs
    ):
    '''
    Info about conditions to test. 

    Parameters
    ----------
    channels_list: list
        ...
    condition_names: list
        ...
    conditions_list: list
        ...
    name: str
    validation_prop: float
    n_each: int
    scale: tuple of float
    '''
    # get the kwargs for obtaining the training data
    experiment = {}
    experiment['get_train_data'] = {
       'validation_prop' : validation_prop, 
        'n_each' : n_each, 
        'scale' : scale, 
        'name' : name, 
        'channels' : {}
    }
    for i, nm in enumerate(condition_names):
        experiment['get_train_data']['channels'][nm] = channels_list[i]
    # get the kwargs for training under each condition
    for i in range(len(condition_names)):
        experiment[condition_names[i]] = {
            'scale' : scale,
            'epochs' : 4,
            'lr' : .01,
            'loss_function' : 'BCELoss',
            'chan_weights' : None, 
            'weights' : None,
            'update_every' : 20, 
            'fork_channels' : None
        }
        if conditions_list is not None:
            custom_kw = conditions_list[i]
            for key in custom_kw.keys():
                experiment[condition_names[i]][key] = custom_kw[key]
    if 'mask' in experiment['get_train_data']['channels']:
        experiment['get_train_data']['absolute_thresh'] = 0.5
    return experiment


# -----------------
# Experiments
# -----------------

lsr_exp = get_experiment_dict(
    [('z-1-smooth', 'y-1-smooth', 'x-1-smooth', 'mask', 'centreness-log'), 
     ('z-1', 'y-1', 'x-1', 'mask', 'centreness-log')], 
    ['z-1s_y-1s_x-1s_m_cl', 'z-1_y-1_x-1_m_cl'], 
    name='label-smoothing-reg-exp'
)

affinities_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness-log'), 
     ('z-1', 'z-2', 'y-1', 'y-2', 'x-1', 'x-2', 'mask', 'centreness-log'),
     ('z-1', 'z-3', 'z-2', 'y-1', 'y-2', 'y-3', 'x-1', 'x-2', 'x-3', 'mask', 'centreness-log')], 
    ['z-1_y-1_x-1_m_cl', 'z-1_z-2_y-1_y-2_x-1_x-2_m_cl', 'z-1_z-2_z-3_y-1_y-2_y-3_x-1_x-2_x-3_m_cl'], 
     name='affinities-exp'
)

thresh_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness-log'), 
     ('z-1', 'y-1', 'x-1', 'centreness', 'centreness-log')], 
    ['z-1_y-1_x-1_m_cl', 'z-1_y-1_x-1_c_cl'], 
    name='threshold-exp'
)

forked_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness-log'), 
     ('z-1', 'y-1', 'x-1', 'mask', 'centreness-log')], 
    ['z-1_y-1_x-1_m_cl', 'f3,2_z-1_y-1_x-1_m_cl'], 
    [{}, {'fork_channels': (3, 2)}], 
    name ='forked-exp'
)

seed_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness'), 
     ('z-1', 'y-1', 'x-1', 'mask', 'centreness-log'),
     ('z-1', 'y-1', 'x-1', 'mask', 'centroid-gauss')],
    ['z-1_y-1_x-1_m_c', 'z-1_y-1_x-1_m_cl', 'z-1_y-1_x-1_m_cg'], 
    name='seed-exp'
)

loss_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness-log'), 
     ('z-1', 'y-1', 'x-1', 'mask', 'centreness-log')],
    ['BCE_z-1_y-1_x-1_m_cl', 'DICE_z-1_y-1_x-1_m_cl'], 
    [{'loss_function' : 'BCELoss'}, {'loss_function' : 'DICELoss'}], 
    name='loss-exp'
)

lr_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness-log'), 
     ('z-1', 'y-1', 'x-1', 'mask', 'centreness-log'),
     ('z-1', 'y-1', 'x-1', 'mask', 'centreness-log')], 
    ['lr0-05_z-1_y-1_x-1_m_cl', 'lr0-01_z-1_y-1_x-1_m_cl', 'lr0-005_z-1_y-1_x-1_m_cl'], 
    [{'lr' : 0.05}, {'lr' : 0.01}, {'lr' : 0.005}], 
    name='learning-rate-exp'
)

mini_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness-log')], 
    ['z-1_y-1_x-1_m_c'], 
    [{'epochs' : 2}], 
    n_each=25, 
    name='mini-train-unet'
)

basic_exp = get_experiment_dict(
    [('z-1', 'y-1', 'x-1', 'mask', 'centreness-log')], 
    ['z-1_y-1_x-1_m_c'], 
    n_each=50
)

def get_files(dirs, ends='.zarr'):
    files = []
    for d in dirs:
        for sub in os.walk(d):
            if ends.endswith('.zarr'):
                if sub[0].endswith(ends):
                    files.append(sub[0])
            for fl in sub[2]:
                f = os.path.join(sub[0], fl)
                if f.endswith(ends):
                    files.append(f)
    return files

if __name__ == '__main__':
 
    # New code
    out_dir = '/home/abigail/data/plateseg-training/training_output'
    dirs = ['/home/abigail/data/plateseg-training/training_gt/Pia', 
            '/home/abigail/data/plateseg-training/training_gt/Volga', 
            '/home/abigail/data/plateseg-training/training_gt/Abi']
    image_paths = get_files(dirs, ends='_image.zarr')
    gt_paths = get_files(dirs, ends='_labels.zarr')
    unets = run_experiment(basic_exp, image_paths, gt_paths, out_dir)
    


from .augment import augment_images
from datetime import datetime
from .helpers import get_files, log_dir_or_None, write_log, LINE
import numpy as np
import os
import pandas as pd
from pathlib import Path
import re
import skimage.filters as filters
from skimage.measure import regionprops
from skimage.morphology._util import _offsets_to_raveled_neighbors
from tifffile import TiffWriter, imread
from time import time
import torch 
from tqdm import tqdm
import zarr


# -----------------------------------------------------------------------------
# Lable Generating Functions
# -----------------------------------------------------------------------------

def get_training_labels(
                        l, 
                        channels=('z-1', 'y-1', 'x-1', 'centreness'),
                        scale=(4, 1, 1)):
    labels = []
    get_offsets = False
    for chan in channels:
        if chan.startswith('offsets-'):
            get_offsets = True
    if get_offsets:
        offsets = get_centre_offsets(l, scale)
    for chan in channels:
        if chan.startswith('z'):
            axis = 0
        elif chan.startswith('y'):
            axis = 1
        elif chan.startswith('x'):
            axis = 2
        n = re.search(r'\d+', chan)
        if n is not None:
            # get the nth affinity
            n = int(n[0]) 
            lab = nth_affinity(l, n, axis)
        elif chan == 'centreness':
            # get the centreness score
            lab = get_centreness(l, scale=scale)
        elif chan == 'centreness-log':
            lab = get_centreness(l, scale=scale, log=True)
        elif chan == 'centroid-gauss':
            lab = get_gauss_centroids(l)
        elif chan.startswith('offsets-'):
            a = _offset_channel(chan)
            lab = offsets[a]
        elif chan == 'mask':
            lab = get_semantic_labels(l)
        else:
            m = f'Unrecognised channel type: {chan} \n'
            m = m + 'Please enter str of form <axis>-<n> for nth affinity (e.g., z-1), \n'
            m = m + 'centreness for centreness score (option of -log for log of centreness),\n'
            m = m + 'or offset-<axis> (e.g., offset-z) for axis offsets'
            raise ValueError(m)
        if chan.endswith('-smooth'):
            lab = smooth(lab)
        labels.append(lab)
    labels = np.stack(labels, axis=0)
    return labels


def _offset_channel(chan):
    if chan.endswith('z'):
        a = 0
    elif chan.endswith('y'):
        a = 1
    elif chan.endswith('x'):
        a = 2
    else:
        raise ValueError(f'Incompatible offset axis name: {chan}')
    return a


# ----------
# Affinities
# ----------

def nth_affinity(labels, n, axis):
    affinities = []
    labs_pad = np.pad(labels, n, mode='reflect')
    for i in range(labels.shape[axis]):
        s_0 = [slice(None, None)] * len(labs_pad.shape) 
        s_0[axis] = slice(i, i + 1)
        s_0 = tuple(s_0)
        s_n = [slice(None, None)] * len(labs_pad.shape) 
        s_n[axis] = slice(i + n, i + n + 1)
        s_n = tuple(s_n)
        new_0 = labs_pad[s_0]
        new_1 = labs_pad[s_n]
        new = new_0 - new_1
        new = np.squeeze(new)
        if len(new) > 0:
            affinities.append(new)
    affinities = np.stack(affinities, axis=axis)
    s_ = [slice(n, -n)] * len(labs_pad.shape)
    s_[axis] = slice(None, None)
    s_ = tuple(s_)
    affinities = affinities[s_]
    affinities = np.where(affinities != 0, 1., 0.)
    return affinities


# not currently referenced, uses nth_affinity() for generality
def get_affinities(image):
    """
    Get short-range voxel affinities for a segmentation. Affinities are 
    belonging to {0, 1} where 1 represents a segment boarder voxel in a
    particular direction. Affinities are produced for each dimension of 
    the labels and each dim has its own channel (e.g, (3, z, y, x)). 

    Note
    ----
    Others may represent affinities with {-1, 0}, because technically... 
    My network wasn't designed for this :)
    """
    padded = np.pad(image, 1, mode='reflect')
    affinities = []
    for i in range(len(image.shape)):
        a = np.diff(padded, axis=i)
        a = np.where(a != 0, 1.0, 0.0)
        a = a.astype(np.float32)
        s_ = [slice(1, -1)] * len(image.shape)
        s_[i] = slice(None, -1)
        s_ = tuple(s_)
        affinities.append(a[s_])
    affinities = np.stack(affinities)
    return affinities    


# -----------
# Centredness
# -----------

def get_centreness(labels, scale=(4, 1, 1), log=False, power=False):
    """
    Obtains a centreness score for each voxel belonging to a labeled object.
    Values in each object sum to one. Values are inversely proportional
    to euclidian distance from the object centroid.

    Notes
    -----
    Another possible implementation would involve the medioid, as in: 
    Lalit, M., Tomancak, P. and Jug, F., 2021. Embedding-based Instance 
    Segmentation of Microscopy Images. arXiv.

    Unfortunately, skimage doesn't yet have a method for finding the  
    medioid (more dev, *sigh*).
    """
    scale = np.array(scale)
    def dist_score(mask):
        output = np.zeros_like(mask, dtype=np.float32)
        c = np.mean(np.argwhere(mask), axis=0)
        indices, values = inverse_dist_score(
                mask, c, scale, log=log, power=power
                )
        output[indices] = values
        return output
    t = time()
    props = regionprops(labels, extra_properties=(dist_score,))
    new = np.zeros(labels.shape, dtype=np.float32)
    for i, prop in tqdm(enumerate(props), desc='Score centreness'):
        new[prop.slice] += prop.dist_score
    new = np.nan_to_num(new)
    print('------------------------------------------------------------')
    print(f'Obtained centreness scores in {time() - t} seconds')
    return new


def inverse_dist_score(mask, centroid, scale, log, power):
    '''
    Compute euclidian distances of each index from a mask
    representing a single object from the centroid of said object

    Uses scale to account for annisotropy in image
    '''
    indices = np.argwhere(mask > 0)
    distances = []
    centre = centroid
    for i in range(indices.shape[0]):
        ind = indices[i, ...]
        diff = (centre - ind) * scale
        dist = np.linalg.norm(diff)
        if log and abs(dist) > 0:
            m = f'Infinite value with distance of {dist}'
            dist = np.log(dist)
            assert not np.isinf(dist), m
        if power:
            dist = 2 ** dist
        distances.append(dist)
    distances = np.array(distances)
    if log:
        distances = distances + np.abs(distances.min()) # bring min value to 0
    norm_distances = distances / distances.max()
    values = (1 - norm_distances) 
    indices = tuple(indices.T.tolist())
    return indices, values


# --------------
# Centre Offsets
# --------------

def get_centre_offsets(labels, scale):
    m = labels > 0
    m = []
    scale = np.array(scale)
    def offsets(mask):
        shape = np.insert(mask.shape, -3, 3)
        output = np.zeros(shape, dtype=np.float32) 
        #print(output.shape)
        c = np.mean(np.argwhere(mask), axis=0)
        indices, values = centre_offsets(c, mask, scale)
        #print(indices, values)
        output[indices] = values
        return output
    t = time()
    props = regionprops(labels, extra_properties=(offsets,))
    #new = np.zeros(np.insert(labels.shape, -3, 3), dtype=np.float32)
    m = labels > 0
    m = np.stack([m, m.copy(), m.copy()], axis=0)
    new = np.where(m == 1, 0., 0.5)
    for i, prop in tqdm(enumerate(props), desc='Get axial centre offsets'):
        #print(new.shape, prop.offsets.shape, prop.slice)
        s_ = [slice(None, None), ]
        for s in prop.slice:
            s_.append(s)
        s_ = tuple(s_)
        new[s_] += prop.offsets
    new = np.nan_to_num(new)
    print('------------------------------------------------------------')
    print(f'Obtained centre offsets in {time() - t} seconds')
    return new


def centre_offsets(c, mask, scale, axes=3):
    idxs = np.argwhere(mask > 0)
    indices = []
    distances = []
    for a in range(axes):
        a_distances = []
        a_indices = []
        for i in range(idxs.shape[0]):
            idx = idxs[i, ...]
            diff = (c - idx) * scale
            # add indicies and offset
            a_indices.append(np.insert(idx, 0, a))
            a_distances.append(diff[a])
        a_indices = np.array(a_indices)
        indices.append(a_indices)
        a_distances = np.array(a_distances)
        new = []
        for d in a_distances:
            if d > 0:
                new.append((d / a_distances.max()))
            elif d == 0:
                new.append(0)
            elif d < 0:
                new.append(-(d / a_distances.min()))
        d = np.array(new)
        d = d - (-1)
        d = d / 2
        distances.append(d)
    indices = np.concatenate(indices)
    indices = tuple(indices.T.tolist())
    distances = np.concatenate(distances)
    return indices, distances



# ---------------
# Semantic Labels
# ---------------

def get_semantic_labels(labels):
    out = np.where(labels > 1, 1., 0.)
    return out


# ------------------
# Smoothed Centroids
# ------------------

# not used
def get_gauss_centroids(labels, sigma=1, z=0):
    centroids = [prop['centroid'] for prop in regionprops(labels)]
    centroids = tuple(np.round(np.stack(centroids).T).astype(int))
    centroid_image = np.zeros(labels.shape, dtype=float)
    centroid_image[centroids] = 1.
    gauss_cent = []
    for i in range(labels.shape[z]):
        s_ = [slice(None, None)] * labels.ndim
        s_[z] = slice(i, i+1)
        s_ = tuple(s_)
        plane = np.squeeze(centroid_image[s_])
        gauss_cent.append(filters.gaussian(plane, sigma=sigma))
    out = np.stack(gauss_cent, axis=z)
    out = out - out.min()
    out = out / out.max()
    #print(out.dtype, out.shape, out.max(), out.min())
    return out


def smooth(image, z=0, sigma=1):
    out = []
    for i in range(image.shape[z]):
        s_ = [slice(None, None)] * image.ndim
        s_[z] = slice(i, i+1)
        s_ = tuple(s_)
        plane = np.squeeze(image[s_])
        out.append(filters.gaussian(plane, sigma=sigma))
    out = np.stack(out, axis=z)
    return out


# -----------------------------------------------------------------------------
# Log and Print
# -----------------------------------------------------------------------------

def print_labels_info(channels, out_dir=None, log_name='log.txt'):
    print(LINE)
    if isinstance(channels, list):
        s = f'Training labels have {len(channels)} output channels: \n'
        _write_log_0(s, out_dir, log_name)
        print(s)
    if isinstance(channels, dict):
        keys = list(channels.keys())
        n_labs = len(keys)
        s = f'{n_labs} sets of training labels were generated:'
        _write_log_0(s, out_dir, log_name)
        print(s)
        for key in keys:
            s = f'Training labels entitled {key} has {len(channels[key])} output channels:'
            print(s)
            _write_log_0(s, out_dir, log_name)
            _print_chans(channels[key], out_dir, log_name)


def _write_log_0(s, out_dir, log_name):
    if out_dir is not None:
            write_log(LINE, out_dir, log_name)
            write_log(s, out_dir, log_name)


def _print_chans(channels, out_dir, log_name):
    for i, chan in enumerate(channels):
        affinity_match = re.search(r'[xyz]-\d*', chan)
        if affinity_match is not None:
            n = f'{affinity_match[0]} affinities'
        elif chan == 'centreness':
            n = 'centreness score'
        elif chan == 'centreness-log':
            n = 'log centreness score'
        elif chan == 'centroid-gauss':
            n = 'gaussian centroids'
        elif chan.startswith('offsets'):
            a = chan[-1]
            n = f'{a}-axis centre offsets'
        elif chan == 'mask':
            n = 'object mask'
        else:
            n = 'Unknown channel type'
        s = f'Channel {i}: {n}'
        print(s)
        if out_dir is not None:
            write_log(s, out_dir, log_name)


from ._dock_widgets import train_from_viewer, segment_data, combine_layers,  \
    assess_segmentation, compare_segmentations

__all__ = [
    'train_from_viewer', 
    'segment_data', 
    'combine_layers', 
    'generate_ground_truth', 
    'assess_segmentation', 
    'compare_segmentations'
]

__version__ = "0.0.1"




from itertools import permutations
import numpy as np
from skimage.util import random_noise, img_as_float
from skimage.transform import AffineTransform, warp
from zarr.core import Array


def augment_images(image, labels, ground_truth=None, augment_prob=0.9):
    augment = np.random.binomial(1, augment_prob)
    if augment:
        image = augment_intensity(image)
    imgs = [image, ]
    if isinstance(labels, dict):
        for key in labels.keys():
            imgs.append(labels[key])
    else:
        imgs.append(labels)
    if ground_truth is not None:
        imgs.append(ground_truth)
    if augment:
        imgs = augment_order(imgs)
    result = [imgs[0], ]
    if isinstance(labels, dict):
        keys = list(labels.keys())
        labs = {key : imgs[i + 1] for i, key in enumerate(keys)}
    else:
        labs = imgs[1]
    result.append(labs)
    if ground_truth is not None:
        gt = imgs[-1]
        result.append(gt)
    return tuple(result)


def augment_intensity(
            image, 
            min_shift=-0.1, 
            max_shift=0.1, 
            min_scale=0.8, 
            max_scale=1.2, 
            shift_sigma=0.02, 
            scale_sigma=.05, 
            noise_prob=0.3, 
            verbose=False
        ):
    if isinstance(image, Array):
        image = np.array(image)
    out = image.copy() / image.max()
    # Get the intensity scale
    scale = continuous_choice(min_scale, max_scale, scale_sigma, loc=1.)
    # Get the intensity shift
    shift = continuous_choice(min_shift, max_shift, shift_sigma)
    out = (out * scale) + shift 
    # if add noise is chosen, add some type of noise to the image
    add_noise = np.random.binomial(1, noise_prob)
    if add_noise:
        options = ['gaussian', 'localvar', 'poisson', 'speckle', 'gaussian', 'speckle']
        i = np.random.randint(len(options))
        mode = options[i]
        kwargs = {}
        if mode in ['gaussian', 'speckle']:
            kwargs['var'] = 0.001
        if verbose:
            print(f'adding {mode} noise')
        out = random_noise(out, mode=mode, **kwargs)
    else:
        # clip the brightness range between 0-1. 
        # only necessary if noise isnt added because the noise function 
        # will do this for us
        out = np.where(out < 0, 0, out)
        out = np.where(out > 1, 1, out)
    return out


def augment_order(
            images, # need to take a list of images to support image and labels
            mirror_prob=0.2, 
            transpose_prob=0.2, 
            used_axes=(-2, -1), # must be in the order they appear in original image shape
            verbose=False
        ):
    out = [_copy_for_zarr(image) for image in images]
    new_out = []
    mirror = np.random.binomial(1, mirror_prob)
    if mirror:
        i = np.random.randint(0, len(used_axes))
        axis = used_axes[i]
        if verbose:
            print('mirroring along ', axis)
    for img in out:   
        s_ = [slice(None, None)] * len(img.shape)
        if mirror:
            s_[axis] = slice(None, None, -1)
        img = img[tuple(s_)]
        new_out.append(img)
    out = new_out
    # get the axis permulations for each image
    transpose = np.random.binomial(1, transpose_prob)
    if transpose:
        ps = permutations(used_axes)
        ps = [p for p in ps if p != tuple(used_axes)]
        idx = np.random.randint(0, len(ps))
        p = ps[idx]
    new_out = []
    for image in out:
        # apply permuatations randomly
        if transpose: 
            axes = [i for i in range(image.ndim)]
            for i, ax in enumerate(used_axes):
                na = p[i]
                if na < 0: # convert negative indices to positive
                    na = len(axes) + na
                axes[ax] = na
            if verbose:
                print('transposing to: ', axes)
            image = np.transpose(image, axes)
        new_out.append(image)
    return new_out


def _copy_for_zarr(image):
    if isinstance(image, Array):
        image = np.array(image) 
    return image.copy()


def augment_affine(
        image, 
        min_scale=0.9, 
        max_scale=1.1, 
        scale_sigma=0.05,
        min_rotation=None, 
        max_rotation=None, 
        rotation_sigma=None,
        min_shear=None,
        max_shear=None,
        shear_sigma=None,
        min_translation=None,
        max_translation=None, 
        translation_sigma=None, 
    ):
    """
    Need to calculate where image edges will be!!
    """
    
    scale = continuous_choice(max_scale, min_scale, scale_sigma, loc=1.)
    rotation = continuous_choice(max_rotation, min_rotation, rotation_sigma)
    shear = continuous_choice(max_shear, min_shear, shear_sigma)
    translation = continuous_choice(max_translation, min_translation, translation_sigma)
    # assumes axis 0 is z. We don't want to distort z due to anisotropy
    tform = AffineTransform(
            scale=scale, 
            rotation=rotation, 
            shear=shear, 
            translation=translation
        )
    # find the required slice of image for affine
    bs_ = ...
    # find the slice required to get correct sized data from image
    ns_ = ...
    # get the image
    out = []
    for i in range(image.shape[0]):
        plane = warp(image[i, ...], tform.inverse)
        out.append(plane)
    out = np.stack(out)



def continuous_choice(min_, max_, sigma, loc=0., size=1):
    '''
    Choose a continuous augmentation parameter from between a selected range.
    Parameter choice is approximately normally distributed (if truncated)
    with std dev 'sigma' and mean 'loc'.
    '''
    done = False
    c = 0
    while not done:
        out = np.random.normal(loc=loc, scale=sigma, size=size)
        if size == 1:
            val = out
        else:
            val = out.mean()
        done = val >= min_ and val <= max_
    return out


if __name__ == '__main__':
    import os
    import zarr
    import napari
    data_dir = '/Users/amcg0011/Data/pia-tracking/cang_training'
    img = zarr.open(os.path.join(data_dir, '191113_IVMTR26_I3_E3_t58_cang_training_image.zarr'))
    lab = zarr.open(os.path.join(data_dir, '191113_IVMTR26_I3_E3_t58_cang_training_labels.zarr'))
    v = napari.Viewer()
    for i in range(5):
        print(i)
        #nimg = augment_intensity(img)
        #nimg = augment_order([nimg])[0]
        nimg, nlab = augment_images(img, lab)
        v.add_image(nimg, visible=False)
        v.add_labels(nlab, visible=False)
    napari.run()
    




from heapq import heappop, heappush
from collections import namedtuple

from scipy import ndimage as ndi
import numba
import numpy as np
from skimage import filters, morphology, feature
from skimage.morphology._util import (
        _offsets_to_raveled_neighbors, _validate_connectivity
        )

# ---------
# Watershed
# ---------


def affinity_watershed(
        image, marker_coords, mask, scale=None, out=None, py_func=False
        ):
    image_raveled, marker_coords, offsets, mask, output = _prep_data(
            image, marker_coords, mask, output=out
            )
    if scale is not None:
        image_raveled *= np.abs(scale)
    if py_func:
        raveled_affinity_watershed.py_func(
                image_raveled, marker_coords, offsets, mask, output
                )
    else:
        raveled_affinity_watershed(
                image_raveled, marker_coords, offsets, mask, output
                )
    shape = image.shape[1:]
    output = output.reshape(shape)
    return output


def _prep_data(image, marker_coords, mask=None, output=None):
    # INTENSITY VALUES
    im_ndim = image.ndim - 1  # the first dim should represent affinities
    image_shape = image.shape[1:]
    raveled_image = np.zeros((image.shape[0], image[0].size),
                             dtype=image.dtype)
    for i in range(image.shape[0]):
        raveled_image[i] = image[i].ravel()
    # NEIGHBORS
    selem, centre = _validate_connectivity(im_ndim, 1, None)
    # array of shape (ndim * 2, 2) giving the indicies of neighbor affinities
    offsets = _indices_to_raveled_affinities(image_shape, selem, centre)
    raveled_markers = np.apply_along_axis(
            _raveled_coordinate, 1, marker_coords, shape=image_shape
            )
    if mask is None:
        small_shape = [s - 2 for s in image_shape]
        mask = np.ones(small_shape, dtype=bool)
        mask = np.pad(mask, 1, constant_values=0)
        assert image_shape == mask.shape
    mask_raveled = mask.ravel()
    if output is None:
        output = np.zeros(mask_raveled.shape, dtype=np.int32)
    labels = np.arange(len(raveled_markers)) + 1
    output[raveled_markers] = labels
    return (raveled_image, raveled_markers, offsets, mask_raveled, output)


def _raveled_coordinate(coordinate, shape):
    # array[z, y, x] = (
    #     array.ravel()[
    #         z * array.shape[1] * array.shape[2]
    #         + y * array.shape[2]
    #         + x
    #     ]
    # )
    raveled_coord = 0
    for i in range(len(coordinate)):
        to_add = coordinate[i]
        for j in range(len(shape)):
            if j > i:
                to_add *= shape[j]
        raveled_coord += to_add
    return raveled_coord


def _indices_to_raveled_affinities(image_shape, selem, centre):
    im_offsets = _offsets_to_raveled_neighbors(image_shape, selem, centre)
    # im_offsets[-len(image_shape):] = 0
    affs = np.concatenate([
            np.arange(len(image_shape)),
            np.arange(len(image_shape))[::-1]
            ])
    indices = np.stack([affs, im_offsets], axis=1)
    return indices


@numba.jit(cache=True, nopython=False)
def raveled_affinity_watershed(
        image_raveled, marker_coords, offsets, mask, output
        ):
    """Compute affinity watershed on raveled arrays.

    Parameters
    ----------
    image_raveled : 2D array of float(32), shape (npixels, ndim)
        The z, y, and x affinities around each pixel.
    marker_coords : 1D array of int
        The location of each marker along the pixels dimension of
        ``image_raveled``.
    offsets : 1D array of int
        The signed offsets to each neighboring pixel.
    mask : 1D array of bool, shape (npixels,)
        True for pixels to which the watershed should spread.
    output : 1D array of int
        The output array for markers.
    """
    n_neighbors = offsets.shape[0]
    age = 0
    marker_coords = marker_coords.astype(np.intp)
    offsets = offsets.astype(np.intp)
    aff_offsets = offsets.copy().astype(np.intp)
    aff_offsets[:int(len(offsets) / 2), 1] = 0
    heap = [
            Element(
                    image_raveled[0, 0], age, marker_coords[0],
                    marker_coords[0]
                    )
            ]
    _ = heappop(heap)
    # add each seed to the stack
    for i in range(marker_coords.shape[0]):
        index = marker_coords[i]
        value = np.float32(0.)
        source = index
        index = index
        elem = Element(value, age, index, source)
        heappush(heap, elem)
    # remove from stack until empty
    while len(heap) > 0:
        elem = heappop(heap)
        for i in range(n_neighbors):
            # get the flattened address of the neighbor
            # offsets are 2d (size, 2) with columns 0 and 1 corresponding to
            # affinities (ie axis) and image neighbour indices respectively
            neighbor_index = elem.index + offsets[i, 1]
            if not mask[neighbor_index]:
                # neighbor is not in mask, move on to next neighbor
                continue
            if output[neighbor_index]:
                # if there is a non-zero value in output, move on to next
                # neighbor
                continue
            # if the neighbor is in the mask and not already labeled,
            # label it then add it to the queue
            output[neighbor_index] = output[elem.index]
            value = image_raveled[aff_offsets[i, 0],
                                  aff_offsets[i, 1] + elem.index]
            age += 1
            new_elem = Element(value, age, neighbor_index, elem.source)
            heappush(heap, new_elem)
    return output


Element = namedtuple('Element', ['value', 'age', 'index', 'source'])


def segment_output_image(
        unet_output,
        affinities_channels,
        centroids_channel,
        thresholding_channel,
        scale=None,
        absolute_thresh=None,
        out=None,
        py_func=False,
        ):
    '''
    Parameters
    ----------
    unet_output: np.ndarray or dask.array.core.Array
        Output from U-net inclusive of all channels. If there is an extra
        dim of size 1, this will be squeezed out. Therefore shape may be
        (1, c, z, y, x) or (c, z, y, x).
    affinities_channels: tuple of int
        Ints, in order (z, y, x) describe the channel indicies to which
        the z, y, and x short-range affinities belong.
    centroids_channel: int
        Describes the channel index for the channel that is used to find
        centroids.
    thresholding_channel: in
        Describes the channel index for the channel that is used to find
        the mask for watershed.
    '''
    unet_output = np.asarray(np.squeeze(unet_output))
    # Get the affinities image (a, z, y, x)
    affinities = unet_output[list(affinities_channels)]
    affinities /= np.max(affinities, axis=(1, 2, 3)).reshape((-1, 1, 1, 1))
    affinities = np.pad(
            affinities,
            ((0, 0), (1, 1), (1, 1), (1, 1)),
            mode='constant',
            constant_values=0,
            )
    # Get the image for finding centroids
    centroids_img = unet_output[centroids_channel]
    # find the centroids
    centroids = _get_centroids(centroids_img) + 1  # account for padding
    # Get the image for finding the mask
    masking_img = unet_output[thresholding_channel]
    # find the mask for use with watershed
    if absolute_thresh is None:
        mask = _get_mask(masking_img)
    else:
        mask = masking_img > absolute_thresh
    mask = np.pad(mask, 1, constant_values=0)  # edge voxels must be 0
    mask, centroids = _remove_unwanted_objects(
            mask, centroids, min_area=10, max_area=10000000
            )
    # affinity-based watershed
    segmentation = affinity_watershed(
            affinities, centroids, mask, scale=scale, out=out, py_func=py_func
            )
    segmentation = segmentation[1:-1, 1:-1, 1:-1]
    seeds = centroids - 1
    return segmentation, seeds, mask


def _get_mask(img, sigma=2):
    thresh = filters.threshold_otsu(filters.gaussian(img, sigma=sigma))
    mask = img > thresh
    return mask


def _get_centroids(cent, gaussian=True):
    if gaussian:
        cent = filters.gaussian(cent, sigma=(0, 1, 1))
    centroids = feature.peak_local_max(cent, threshold_abs=.04)
    return centroids


def _remove_unwanted_objects(mask, centroids, min_area=0, max_area=1000000):
    labels, _ = ndi.label(mask)
    labels_no_small = morphology.remove_small_objects(
            labels, min_size=min_area
            )
    labels_large = morphology.remove_small_objects(
            labels_no_small, min_size=max_area
            )
    labels_goldilocks = labels_no_small ^ labels_large
    centroid_labels = labels_goldilocks[tuple(centroids.T)]
    new_centroids = centroids[centroid_labels > 0]
    new_mask = labels_goldilocks.astype(bool)
    return new_mask, new_centroids


if __name__ == '__main__':
    from skimage import data, segmentation, morphology
    foreground = data.binary_blobs(length=128, n_dim=3, volume_fraction=0.35)
    centroids = ndi.distance_transform_edt(foreground)
    affz, affy, affx = [
            filters.scharr(foreground.astype(float), axis=i) for i in range(3)
            ]
    volume = np.stack([affz, affy, affx, centroids, foreground],
                      axis=0).astype(np.float32)
    import time
    start = time.time()
    _, _, _ = segment_output_image(
            volume, (0, 1, 2), 3, 4, absolute_thresh=0.5
            )
    end = time.time()
    print('warmup', end - start)

    start = time.time()
    labels, _, _ = segment_output_image(
            volume, (0, 1, 2), 3, 4, absolute_thresh=0.5
            )
    end = time.time()
    print('second run', end - start)

    start = time.time()
    labels_py, _, _ = segment_output_image(
            volume, (0, 1, 2), 3, 4, absolute_thresh=0.5, py_func=True
            )
    end = time.time()
    print('python', end - start)

    start = time.time()
    _ = segmentation.watershed(
            np.sqrt(np.sum(volume[:3]**2, axis=0)),
            ndi.label(morphology.local_maxima(centroids, indices=False))[0],
            mask=foreground,
            )
    end = time.time()
    print('skimage watershed', end - start)

    print('numba and python are equal', np.all(labels == labels_py))

    import napari
    napari.view_labels(labels, ndisplay=3)
    napari.run()



from os import path
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
from pathlib import Path
#import ptitprince as pt
import seaborn as sns
from typing import Union
import matplotlib

# ----------
# Loss Plots
# ----------

def save_loss_plot(path, loss_function, v_path=None, show=True):
    df = pd.read_csv(path)
    if v_path is not None:
        vdf = pd.read_csv(v_path)
    else:
        vdf = None
    p = Path(path)
    n = p.stem
    d = p.parents[0]
    out_path = os.path.join(d, n + '_loss.png')
    fig, ax = plot_loss(df, vdf=vdf, x_lab='Iteration', y_lab=loss_function, save=out_path, show=show)



def plot_loss(df, vdf=None, x_lab='Iteration', y_lab='BCE Loss', save=None, show=True):
    x = df['Unnamed: 0'].values
    y = df['loss'].values
    epochs = len(df['epoch'].unique())
    no_batches = int(len(x) / epochs)
    epoch_ends = np.array([((i + 1) * no_batches) - 1 for i in range(epochs)])
    epoch_end_x = x[epoch_ends]
    epoch_end_y = y[epoch_ends]
    fig, ax = plt.subplots()
    leg = ['loss',]
    ax.plot(x, y, linewidth=2)
    ax.scatter(epoch_end_x, epoch_end_y)
    title = 'Training loss'
    if vdf is not None:
        if len(vdf) > epochs:
            vy = vdf.groupby('batch_id').mean()['validation_loss'].values
            vx = vdf['batch_id'].unique()
        else:
            vy = vdf['validation_loss'].values
            vx = epoch_end_x
        title = title + ' with validation loss'
        leg.append('validation loss')
        if len(vdf) > epochs:
            #vy_err = v_df.groupby('batch_id').sem()['validation_loss'].values
            #ax.errorbar(vx, vy, vy_err, marker='.')
            ax.plot(vx, vy, linewidth=2, marker='o')
        else:
            ax.plot(vx, vy, linewidth=2, marker='o')
    ax.set(xlabel=x_lab, ylabel=y_lab)
    ax.set_title(title)
    ax.legend(leg)
    fig.set_size_inches(13, 9)
    if save is not None:
        plt.savefig(save, dpi=300)
    if show:
        plt.show()
    return fig, ax



def save_channel_loss_plot(path, show=True):
    df = pd.read_csv(path)
    p = Path(path)
    n = p.stem
    d = p.parents[0]
    out_path = os.path.join(d, n + '_channel-loss.png')
    fig, ax = plot_channel_losses(df, save=out_path, show=show)



def plot_channel_losses(df, x_lab='Iteration', y_lab='BCE Loss', save=None, show=True):
    cols = list(df.columns)
    x = df['Unnamed: 0'].values
    non_channel_cols = ['Unnamed: 0', 'epoch', 'batch_num', 'loss', 'data_id']
    channel_losses = [col for col in cols if col not in non_channel_cols]
    #print(channel_losses)
    if len(channel_losses) > 5:
        #print('four plots')
        fig, axs = plt.subplots(2, 2)
        zs, ys, xs, cs = [], [], [], []
        for col in channel_losses:
            y = df[col].values
            if col.startswith('z'):
                ls = _get_linestyle(zs)
                axs[0, 0].plot(x, y, linewidth=1, linestyle=ls)
                zs.append(col)
            if col.startswith('y'):
                ls = _get_linestyle(ys)
                axs[0, 1].plot(x, y, linewidth=1, linestyle=ls)
                ys.append(col)
            if col.startswith('x'):
                ls = _get_linestyle(xs)
                axs[1, 0].plot(x, y, linewidth=1, linestyle=ls)
                xs.append(col)
            if col.startswith('cent') or col == 'mask':
                ls = _get_linestyle(cs)
                axs[1, 1].plot(x, y, linewidth=1, linestyle=ls)
                cs.append(col)
        axs[0, 0].set_title('Z affinities losses')
        axs[0, 0].legend(zs)
        axs[0, 1].set_title('Y affinities losses')
        axs[0, 1].legend(ys)
        axs[1, 0].set_title('X affinities losses')
        axs[1, 0].legend(xs)
        axs[1, 1].set_title('Object interior losses')
        axs[1, 1].legend(cs)
        fig.set_size_inches(13, 9)
    elif len(channel_losses) <= 5:
        #print('two plots')
        fig, axs = plt.subplots(2, 1)
        affs, cs = [], []
        for col in channel_losses:
            y = df[col].values
            if col.startswith('z') or col.startswith('y') or col.startswith('x'):
                ls = _get_linestyle(affs)
                axs[0].plot(x, y, linewidth=2, linestyle=ls)
                affs.append(col)
            if col.startswith('cent') or col == 'mask':
                axs[1].plot(x, y, linewidth=2)
                cs.append(col)
        axs[0].set_title('Affinities losses')
        axs[0].legend(affs)
        axs[1].set_title('Object interior losses')
        axs[1].legend(cs)
        fig.set_size_inches(14, 14)
    for ax in axs.flat:
        ax.set(xlabel=x_lab, ylabel=y_lab)
    if save is not None:
        plt.savefig(save, dpi=300)
    if show:
        plt.show()
    return fig, axs


def _get_linestyle(lis):
    if len(lis) == 0:
        ls = '-'
    elif len(lis) == 1:
        ls = '--'
    else:
        ls = ':'
    return ls


# --------
# VI Plots
# --------


def VI_plot(
            df, 
            cond_ent_over="GT | Output",  
            cond_ent_under="Output | GT", 
            lab='Variation of information',
            save=False, 
            show=True, 
            ax=None, 
            title=True, 
            palette='Set2', 
            orient='h', 
            sigma=0.2, 
            compare=False
            ):
    #df = pd.read_csv(path)
    overseg = df[cond_ent_over].values
    o_groups = [cond_ent_over] * len(overseg)
    underseg = df[cond_ent_under].values
    u_groups = [cond_ent_under] * len(underseg)
    groups = o_groups + u_groups
    x = lab
    y = 'Conditional entropy'
    data = {
        x : groups, 
        y : np.concatenate([overseg, underseg])
        }
    data = pd.DataFrame(data)
    if ax is None:
        f, ax = plt.subplots(figsize=(8, 6))
    
    sns.boxplot(x=x, y=y, data=data, palette=palette, ax=ax)
    sns.stripplot(x=x, y=y, data=data, palette=palette, edgecolor="white", ax=ax,
                 size=3, jitter=1, zorder=0, orient=orient,  dodge=True, linewidth=0.3)
    sns.boxplot()
    #pt.RainCloud(x=x, y=y, data=data, palette=palette, bw=sigma,
    #             width_viol=.6, ax=ax, orient=orient)
    p = Path(save)
    if title:
        plt.title(p.stem)
    if save:
        plt.savefig(save, bbox_inches='tight')
    if show:
        plt.show()



def VI_plot_compare(
            df, 
            ax0,
            ax1, 
            comparison_name,
            conditions,
            cond_ent_over="VI: GT | Output",  
            cond_ent_under="VI: Output | GT", 
            palette='Set2', 
            orient='h', 
            sigma=0.2,
            name='model_name',
            ):
    sns.boxplot(x=name, y=cond_ent_over, data=df, palette=palette, ax=ax0)
    sns.stripplot(x=name, y=cond_ent_over, data=df, palette=palette, edgecolor="white", ax=ax0,
                 size=3, jitter=1, zorder=0, orient=orient,  dodge=True, linewidth=0.3)
    #pt.RainCloud(x=name, y=cond_ent_over, hue=name, hue_order=conditions, data=df, palette=palette, bw=sigma,
     #            width_viol=.6, ax=ax0, orient=orient, alpha=0.8, move=0.3)
    ax0.set_ylabel(comparison_name)
    sns.despine(ax=ax0)
    ax0.legend([],[], frameon=False)
    sns.boxplot(x=name, y=cond_ent_under, data=df, palette=palette, ax=ax1)
    sns.stripplot(x=name, y=cond_ent_under, data=df, palette=palette, edgecolor="white", ax=ax1,
                 size=3, jitter=1, zorder=0, orient=orient,  dodge=True, linewidth=0.3)
    #pt.RainCloud(x=name, y=cond_ent_under, hue=name, hue_order=conditions, data=df, palette=palette, bw=sigma,
    #             width_viol=.6, ax=ax1, orient=orient, alpha=0.8, move=0.3)
    ax1.set_ylabel(comparison_name)
    sns.despine(ax=ax1)
    ax1.legend([],[], frameon=False)



def experiment_VI_plots(
        dfs, 
        names, 
        title,
        out_name,
        out_dir,
        cond_ent_over="GT | Output",  
        cond_ent_under="Output | GT", 
        show=True
    ):
    plt.rcParams.update({'font.size': 16})
    groups = []
    ce0 = []
    ce1 = []
    for i, df in enumerate(dfs):
        ce0.append(df[cond_ent_over].values)
        ce1.append(df[cond_ent_under].values)
        groups += [names[i]] * len(df)
    x = 'Experiment'
    data = {
        x : groups, 
        cond_ent_over : np.concatenate(ce0), 
        cond_ent_under : np.concatenate(ce1)
    }
    data = pd.DataFrame(data)
    o = 'h'
    pal = 'Set2'
    sigma = .2
    f, axs = plt.subplots(1, 2, figsize=(8, 6)) #, sharex=True) #, sharey=True)
    ax0 = axs[0]
    ax1 = axs[1]
    sns.boxplot(x=x, y=cond_ent_over, data=data, palette=pal, ax=ax0)
    sns.stripplot(x=x, y=cond_ent_over, data=data, palette=pal, edgecolor="white", ax=ax0,
                 size=3, jitter=1, zorder=0, orient=o, dodge=True, linewidth=0.3)
    #pt.RainCloud(x = x, y = cond_ent_over, data = data, palette = pal, bw = sigma,
    #             width_viol = .6, ax = ax0, orient = o, move=0.3)
    ax0.set_title('Over-segmentation conditional entropy')
    sns.boxplot(x=x, y=cond_ent_under, data=data, palette=pal, ax=ax0)
    sns.stripplot(x=x, y=cond_ent_under, data=data, palette=pal, edgecolor="white", ax=ax0,
                 size=3, jitter=1, zorder=0, orient=o, dodge=True, linewidth=0.3)
    #pt.RainCloud(x = x, y = cond_ent_under, data = data, palette = pal, bw = sigma,
    #             width_viol = .6, ax = ax1, orient = o, move=0.3)
    ax1.set_title('Under-segmentation conditional entropy')
    f.suptitle(title)
    os.makedirs(out_dir, exist_ok=True)
    save_path = os.path.join(out_dir, out_name + '_VI_rainclould_plots.png')
    plt.savefig(save_path, bbox_inches='tight')
    if show:
        plt.show()



# -----------------------
# Average Precision Plots
# -----------------------

def plot_experiment_APs(paths, names, title, out_dir, out_name, show=True):
    dfs = [pd.read_csv(path) for path in paths]
    os.makedirs(out_dir, exist_ok=True)
    out_path = os.path.join(out_dir, out_name)
    plot_AP(dfs, names, out_path, title, show=show)



def plot_AP(dfs, names, out_path, title, 
            thresh_name='threshold', ap_name='average_precision', 
            show=True, add_title=True, ):
    plt.rcParams.update({'font.size': 16})
    plt.rcParams["figure.figsize"] = (8,8)
    fig = plt.figure()
    for df in dfs:
        plt.plot(df[thresh_name].values, df[ap_name].values)
    plt.xlabel('IoU threshold')
    plt.ylabel('Average precision')
    if add_title:
        plt.title(title)
    plt.legend(names)
    fig.savefig(out_path)
    if show:
        plt.show()


# ------------------------------
# Object Number Difference Plots
# ------------------------------

def plot_experiment_no_diff(paths, names, title, out_dir, out_name, col_name='n_diff', show=True):
    dfs = [pd.read_csv(path) for path in paths]
    plt.rcParams.update({'font.size': 16})
    os.makedirs(out_dir, exist_ok=True)
    out_path = os.path.join(out_dir, out_name)
    groups = []
    n_diff = []
    for i, df in enumerate(dfs):
        vals = df[col_name].values
        n_diff.append(vals)
        groups += [names[i]] * len(df)
    x = 'Experiment'
    data = {
        x : groups, 
        'n_diff' : np.concatenate(n_diff), 
    }
    data = pd.DataFrame(data)
    o = 'h'
    pal = 'Set2'
    sigma = .2
    f, ax = plt.subplots(figsize=(10, 10))
    sns.boxplot(x=x, y='n_diff', data=data, palette=pal, ax=ax)
    sns.stripplot(x=x, y='n_diff', data=data, palette=pal, edgecolor="white", ax=ax,
                 size=3, jitter=1, zorder=0, orient=o, dodge=True, linewidth=0.3)
    #pt.RainCloud(x=x, y='n_diff', data=data, palette=pal, bw=sigma,
    #             width_viol=.6, ax=ax, orient=o)
    plt.title(title)
    f.savefig(out_path)
    if show:
        plt.show()
    

def plot_count_difference(
        df, 
        title, 
        out_path, 
        col_name='Count difference', 
        show=True, 
        ):
    plt.rcParams.update({'font.size': 16})
    groups = ['model', ] * len(df)
    n_diff = df[col_name].values
    #for i, df in enumerate(dfs):
     #   vals = df[col_name].values
      #  n_diff.append(vals)
       # groups += [names[i]] * len(df)
    x = 'Experiment'
    data = {
        x : groups, 
        'n_diff' : n_diff, 
    }
    data = pd.DataFrame(data)
    o = 'h'
    pal = 'Set2'
    sigma = .2
    f, ax = plt.subplots(figsize=(10, 10))
    sns.boxplot(x=x, y='n_diff', data=data, palette=pal, ax=ax)
    sns.stripplot(x=x, y='n_diff', data=data, palette=pal, edgecolor="white", ax=ax,
                 size=3, jitter=1, zorder=0, orient=o, dodge=True, linewidth=0.3)
    #pt.RainCloud(x=x, y='n_diff', data=data, palette=pal, bw=sigma,
     #            width_viol=.6, ax=ax, orient=o)
    plt.title(title)
    f.savefig(out_path)
    if show:
        plt.show()



def compare_count_difference(
        df, 
        ax,
        comparison_name,
        conditions, 
        col_name='Count difference', 
        palette='Set2', 
        orient='h', 
        sigma=0.2,
        name='model_name',
        ):
    conditions = pd.unique(df[name])
    sns.boxplot(x=name, y=col_name, data=df, palette=palette, ax=ax)
    sns.stripplot(x=name, y=col_name, data=df, palette=palette, edgecolor="white", ax=ax,
                 size=3, jitter=1, zorder=0, orient=orient, dodge=True, linewidth=0.3)
    #pt.RainCloud(x=name, y=col_name, hue=name, hue_order=conditions, data=df, palette=palette, bw=sigma,
    #             width_viol=.6, ax=ax, orient=orient, alpha=0.8, move=0.3)
    ax.set_ylabel(comparison_name)
    sns.despine(ax=ax)
    ax.legend([],[], frameon=False)
    


def compare_AP(
        df, 
        ax, 
        palette, 
        conditions,
        name='model_name', 
        ap_col='average_precision', 
        thresh_col='threshold'
        ):
    conditions = pd.unique(df[name])
    sns.lineplot(x=thresh_col, y=ap_col, hue=name, hue_order=conditions, data=df, ax=ax, palette=palette)
    ax.set_xlabel('IOU threshold')
    ax.set_ylabel('Average precision')
    sns.despine(ax=ax)


def comparison_plots(
        comparison_directory: str, 
        save_name: str,
        file_exstention: str ='pdf', 
        output_directory: Union[str, None] = None,
        variation_of_information: bool =True, 
        object_difference: bool =True, 
        average_precision: bool =True, 
        n_rows: int =2, 
        n_col: int =2, 
        comparison_name: str= "Model comparison",
        VI_indexs: tuple =(0, 1), # (0, 0)
        OD_index: int =2, # (0, 1)
        AP_index: int =3, # (1, 0)
        fig_size: tuple =(7, 6), 
        raincloud_orientation: str='h',
        raincloud_sigma: float=0.2,
        palette: str='Set2',
        top_white_space: float =5, #TODO eventually figure out how to make this a slider 0-100
        left_white_space: float =15, 
        right_white_space: float =5, 
        bottom_white_space: float =10, 
        horizontal_white_space: float =40,
        vertical_white_space: float =40,
        font_size: int =30,
        style: str ='ticks', 
        context: str='paper', 
        show: bool=True
    ):
    '''
    Make a custom plot to compare sementation models based on output from 
    the assess segmentation tool. The only requirement is that files you 
    want to compare are in the same directory, which should be specified
    as the comparison_directory. The plots will be saved into a single 
    file <save_name>.<file_exstention> (e.g., )

    Parameters
    ----------
    comparison_directory: str
        Directory in which to look for the data to plot
    save_name:
        Name to give the output file. Please don't add a file
        extension.
    file_exstention: str (.pdf)
        Specify one of the following file types: .pdf, .png, .svg
    output_directory: str or None (None)
        Directory into which to save the file. If None, 
        will save into the comparison directory. 
    variation_of_information: bool (True)
        Should we plot VI? This will be plotted in 2 plots:
        one for oversegmentation and one for undersegmentation. 
    object_difference: bool (True)
        Should we plot OD? Will be plotted into a single plot. 
    average_precision: bool (True)
        Should we plot AP? Will be plotted into a single plot.
    comparison_name: str ("Model comparison")
        Label to give comparison in plots. Will be used as
        an axis label in OD and VI plots
    n_rows: int (2) 
        How many rows of plots. 1 - 4.
        e.g. - two because we need all four plots for VI, OD, and AP
             - four because we need all four plots for VI, OD, and AP
               and want to plot everthing in the same column. 
    n_col: int =2, 
        How many rows of plots. 1 - 4.
        e.g. - two because we need all four plots for VI, OD, and AP
             - four because we need all four plots for VI, OD, and AP
               and want to plot everthing in the same row. 
        (e.g., two because we need all four plots for VI, OD, and AP)
    VI_indexs: tuple of int or int (0, 1)
        Which plot to put the VI plots in. The first index refers to
        the oversegmentation plot and the second refers to the 
        undersegmentation plot For instructions see "Using integer indexes" 
        in the function Notes below. 
    OD_index: tuple of int or int (2)
        Which plot to put the OD plot in.For instructions see 
        "Using integer indexes". 
    AP_index: tuple of int or int (3)
        Which plot to put the AP plot in.For instructions see 
        "Using integer indexes". 
    fig_size: tuple (9, 9)
        Size of the figure you want to make (in inches). 
    raincloud_orientation: str ("h")
        Orientation for raincloudplots. "h" for horizontal. 
        "v" for vertical. 
    raincloud_sigma: float (0.2)
        The sigma value used to construct the kernel density estimate 
        in the raincloudplot. Determines the smoothness of the 
        half violinplot/density estimate parts of the plot. Larger values
        result in smoother curves. 
    palette: str ('Set2')
        pandas colour palette to use. See the pandas palette documentation
        for details. 
    top_white_space: float (3)
        percent of the total figure size that should remain white space 
        above the plots.
    left_white_space: float (15)
        percent of the total figure size that should remain white space 
        to the left of the plots.
    right_white_space: float (5) 
        percent of the total figure size that should remain white space 
        to the right the plots.
    bottom_white_space: float (17)
        percent of the total figure size that should remain white space 
        below the plots.
    horizontal_white_space: float (16)
        percent of the total figure size that should remain white space 
        between the plots horizontally.
    vertical_white_space: float (16)
        percent of the total figure size that should remain white space 
        between the plots vertically.
    font_size: int (12)
        Size of the axis label and ticks font
    style: str ("ticks")
        Pandas style. Please see pandas documentation for more info and
        options.  
    context: str ("paper")
        Pandas context. Please see pandas documentation for more info and
        options.  
    show: bool (True)
        Do you want to see the plot in the matplotlib viewer once it has 
        been saved?

    Notes
    -----

        Using integer indexes
        ---------------------
        Numbers 0-4 that tells you which position to place the 
        oversegmentation and undersegmentation plots, respectively 
        Note that no matter how rows and columns are arranged, 
        the numbering will start at the top left plot and proceed
        left to write (much like reading English).
        e.g. - for VI plots, (0, 1) in a 2x2 grid of plots will place
               the VI plots in top two plots 
             - for OD plot, 3 in a 1x4 grid will place the OD plot 


    '''
    # Read in tables and collate data
    # -------------------------------
    VIOD_files = [os.path.join(comparison_directory, f) \
                  for f in os.listdir(comparison_directory) \
                    if f.endswith('_scores.csv')]
    metrics_VIOD = [pd.read_csv(p) for p in VIOD_files]
    metrics_VIOD = pd.concat(metrics_VIOD).reset_index(drop=True)
    AP_files = [os.path.join(comparison_directory, f) \
                  for f in os.listdir(comparison_directory) \
                    if f.endswith('_AP_curve.csv')]
    metrics_AP = [pd.read_csv(p) for p in AP_files]
    metrics_AP = pd.concat(metrics_AP).reset_index(drop=True)
    
    conditions = pd.unique(metrics_VIOD['model_name'])
    
    # plotting
    # --------
    matplotlib.rcParams.update({'font.size': font_size})
    sns.set_context(context)
    sns.set_style(style)
    plt.rcParams['svg.fonttype'] = 'none' # so the text will be saved with the svg - not curves
    is_int = []
    if variation_of_information:
        is_int.append(isinstance(VI_indexs[0], int))
        is_int.append(isinstance(VI_indexs[1], int))
    if average_precision:
        is_int.append(isinstance(AP_index, int))
    if object_difference:
        is_int.append(isinstance(OD_index, int))
    fig, axs = plt.subplots(nrows=n_rows, ncols=n_col)
    if np.sum(is_int) == len(is_int):
        axs = axs.ravel()
    fig.set_size_inches(fig_size)
    #plt.rcParams.update({'font.size': font_size})
    if variation_of_information:
        VI_plot_compare(metrics_VIOD, axs[VI_indexs[0]], axs[VI_indexs[1]], 
                        comparison_name, conditions,
                        palette=palette, 
                        orient=raincloud_orientation, 
                        sigma=raincloud_sigma, 
                        name='model_name')
    if object_difference:
        compare_count_difference(metrics_VIOD, axs[OD_index], comparison_name, conditions,
                                 col_name='Count difference', 
                                 palette=palette, orient=raincloud_orientation, 
                                 sigma=raincloud_sigma, name='model_name')
    if average_precision:
        compare_AP(metrics_AP, axs[AP_index], palette, conditions, name='model_name', 
                   ap_col='average_precision', thresh_col='threshold')

    # White space
    # -----------
    right = 1 - (right_white_space / 100)
    top = 1 - (top_white_space / 100)
    left = left_white_space / 100
    bottom = bottom_white_space / 100
    wspace = horizontal_white_space / 100
    hspace = vertical_white_space / 100
    fig.subplots_adjust(right=right, left=left, bottom=bottom, top=top, wspace=wspace, hspace=hspace)
    
    matplotlib.rcParams.update({'font.size': font_size})
    
    # Save figure
    # -----------
    if output_directory is None:
        output_directory = comparison_directory
    n = save_name + '.' + file_exstention
    save_path = os.path.join(output_directory, n)
    fig.savefig(save_path)

    if show:
        plt.show()